{******************************************************************************}
{*    "DirectX.    Delphi".                  *}
{*  BHV. www.bhv.ru                                               *}
{*  .. softgl@chat.ru                                                *}
{******************************************************************************}
unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  AppEvnts,
  DirectXGraphics;

type
  TfrmD3D = class(TForm)
    ApplicationEvents1: TApplicationEvents;
    procedure FormCreate(Sender: TObject);
    procedure ApplicationEvents1Minimize(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ApplicationEvents1Restore(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    FD3D            : IDIRECT3D8;
    FD3DDevice      : IDIRECT3DDEVICE8;
    FD3DVB          : IDIRECT3DVERTEXBUFFER8;
    FActive         : BOOL;
    ThisTickCount   : DWORD;
    LastTickCount   : DWORD;
    function Render : HRESULT;
    function InitD3D : HRESULT;
    function InitVB : HRESULT;
    procedure Cleanup;
    procedure ErrorOut (const Caption : PChar; const hError : HRESULT);
  public
    { Public declarations }
  end;

type
  TCUSTOMVERTEX = packed record
    X, Y, Z : Single;
    U, V : Single;
  end;

  TFish = class
    private
      FD3Texture : IDIRECT3DTEXTURE8;
    public
      PosX, PosY, StepX : Single;
      Scale : Single;
      function RotateTexture : HRESULT;
      function Draw : HRESULT;
      constructor Create (const FileName : String; const inR, inG, inB : Byte);
      destructor Destroy; override;
      procedure Move;
  end;

const
  D3DFVF_CUSTOMVERTEX = D3DFVF_XYZ or D3DFVF_TEX1;
  NumFish = 10;

var
  frmD3D: TfrmD3D;
  Frames : Integer = 0;

  Fishes : Array [0..NumFish-1] of TFish;

implementation

{$R *.DFM}

procedure TfrmD3D.Cleanup;
begin
  if Assigned (FD3DVB) then begin
     FD3DVB._Release;
     FD3DVB := nil;
  end;

  if Assigned (FD3DDevice) then begin
     FD3DDevice._Release;
     FD3DDevice := nil;
  end;

  if Assigned (FD3D) then begin
     FD3D._Release;
     FD3D := nil;
  end;
end;

procedure TfrmD3D.ErrorOut (const Caption : PChar; const hError : HRESULT);
begin
  FActive := False;
  Cleanup;
  MessageBox (Handle, PChar(DXGErrorString (hError)), Caption, 0);
end;

function TfrmD3D.InitD3D : HRESULT;
var
  d3ddm : TD3DDISPLAYMODE;
  d3dpp : TD3DPRESENT_PARAMETERS;
  hRet : HRESULT;
begin
  FD3DDevice := nil;
  FD3DVB := nil;
  FD3D := nil;

  FD3D := Direct3DCreate8(D3D_SDK_VERSION);
  if FD3D = nil then begin
     Result := E_FAIL;
     Exit;
  end;

  hRet := FD3D.GetAdapterDisplayMode(D3DADAPTER_DEFAULT, d3ddm);
  if FAILED(hRet) then begin
     Result := hRet;
     Exit;
  end;

  ZeroMemory(@d3dpp, SizeOf(d3dpp));
  with d3dpp do begin
       Windowed := True;
       SwapEffect := D3DSWAPEFFECT_DISCARD;
       BackBufferFormat := d3ddm.Format;
  end;

  Result := FD3D.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, Handle,
                              D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                              d3dpp, FD3DDevice);
end;

function TfrmD3D.Render : HRESULT;
var
  hRet : HRESULT;
  i : Integer;
begin
  if FD3DDevice = nil then begin
     Result := E_FAIL;
     Exit;
  end;

  hRet := FD3DDevice.Clear(0, nil, D3DCLEAR_TARGET, $0000FFFF, 0.0, 0);
  if FAILED(hRet) then begin
     Result := hRet;
     Exit;
  end;

  hRet := FD3DDevice.BeginScene;
  if FAILED(hRet) then begin
     Result := hRet;
     Exit;
  end;

  for i := 0 to NumFish - 1 do Fishes [i].Draw;

  hRet := FD3DDevice.EndScene;
  if FAILED(hRet) then begin
     Result := hRet;
     Exit;
  end;

  Result := FD3DDevice.Present(nil, nil, 0, nil);
end;

procedure TfrmD3D.FormCreate(Sender: TObject);
var
  hRet : HRESULT;
  i : Integer;
begin
  Randomize;
  hRet := InitD3D;
  if Failed (hRet) then ErrorOut ('InitD3D', hRet);
  hRet := InitVB;
  if Failed (hRet) then ErrorOut ('InitVB', hRet);

  for i := 0 to NumFish - 1 do begin
      case random (4) of
        0 : Fishes [i] := TFish.Create ('Fish1.bmp', 0, 255, 0);
        1 : Fishes [i] := TFish.Create ('Fish2.bmp', 255, 0, 0);
        2 : Fishes [i] := TFish.Create ('Fish3.bmp', 0, 255, 0);
        3 : Fishes [i] := TFish.Create ('Fish4.bmp', 0, 255, 0);
      end;
      with Fishes [i] do begin
           PosX := random - 0.5;
           PosY := (random (60) - 30) / 100;
           StepX := (random - 0.5) / 10;
           if StepX < 0 then RotateTexture;
           Scale := (random (60) + 40) / 100;
     end;
  end;
end;

procedure TfrmD3D.ApplicationEvents1Minimize(Sender: TObject);
begin
  FActive := False;
end;

procedure TfrmD3D.FormActivate(Sender: TObject);
begin
  LastTickCount := GetTickCount;
  FActive := True;
end;

procedure TfrmD3D.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
var
  hRet : HRESULT;
  i : Integer;
begin
  if FActive then begin
     Inc (Frames);
     hRet := Render;
     if FAILED(hRet) then begin
        FActive := False;
        ErrorOut ('Render', hRet);
        Exit;
     end;

     ThisTickCount := GetTickCount;
     if ThisTickCount - LastTickCount > 50 then begin
        Caption := 'FPS = ' + Format('%6.2f',
                                 [frames * 1000 / (ThisTickCount - LastTickCount)]);
        Frames := 0;

        for i := 0 to NumFish - 1 do Fishes [i].Move;

        LastTickCount := GetTickCount;
     end;
  end;

  Done := False;
end;

procedure TfrmD3D.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_ESCAPE then Close;
end;

procedure TfrmD3D.ApplicationEvents1Restore(Sender: TObject);
begin
  FActive := True;
end;

procedure TfrmD3D.FormDestroy(Sender: TObject);
var
  i : Integer;
begin
  for i := 0 to NumFish - 1 do
      if Assigned (Fishes [i]) then Fishes [i].Free;
  Cleanup
end;

function TfrmD3D.InitVB : HRESULT;
var
  hRet : HRESULT;
begin
  hRet := FD3DDevice.CreateVertexBuffer(4 * SizeOf(TCustomVertex), 0,
                                        D3DFVF_CUSTOMVERTEX,
                                        D3DPOOL_DEFAULT, FD3DVB);
  if Failed(hRet) then begin
     Result := hRet;
     Exit;
  end;

  hRet := FD3DDevice.SetStreamSource(0, FD3DVB, SizeOf(TCustomVertex));
  if Failed(hRet) then begin
     Result := hRet;
     Exit;
  end;

  Result := FD3DDevice.SetVertexShader(D3DFVF_CUSTOMVERTEX);
end;

constructor TFish.Create (const FileName : String; const inR, inG, inB : Byte);
var
  hRet : HRESULT;
  d3dlr : TD3DLOCKED_RECT;
  dwDstPitch : DWORD;
  X, Y : DWORD;
  Bmp, wrkBmp : TBitmap;
  R, G, B : Byte;
begin
  Bmp := TBitmap.Create;
  Bmp.LoadFromfile (FileName);

  wrkBmp := TBitmap.Create;
  wrkBmp.Width := 128;
  wrkBmp.Height := 128;
  wrkBmp.Canvas.StretchDraw (Rect (0, 0, 128, 128), Bmp);

  hRet := frmD3D.FD3DDevice.CreateTexture (wrkBmp.Width, wrkBmp.Height, 0, 0,
                                    D3DFMT_A8R8G8B8,
                                    D3DPOOL_MANAGED, FD3Texture);
  if FAILED(hRet) then begin
     if Failed (hRet) then frmD3D.ErrorOut ('InitTexture', hRet);
     Exit;
  end;

  hRet := FD3Texture.LockRect(0, d3dlr, nil, 0);
  if FAILED(hRet) then begin
     if Failed (hRet) then frmD3D.ErrorOut ('InitTexture', hRet);
     Exit;
  end;

  dwDstPitch := d3dlr.Pitch;
  for Y := 0 to wrkBmp.Height - 1 do
      for X := 0 to wrkBmp.Width - 1 do begin
         R := GetRValue (wrkBmp.Canvas.Pixels [X, DWORD (wrkBmp.Height - 1) - Y]);
         G := GetGValue (wrkBmp.Canvas.Pixels [X, DWORD (wrkBmp.Height - 1) - Y]);
         B := GetBValue (wrkBmp.Canvas.Pixels [X, DWORD (wrkBmp.Height - 1) - Y]);

         if (R = inR) and (G = inG) and (B = inB)
         then PDWORD (DWORD(d3dlr.pBits) + Y * dwDstPitch + X * 4)^ := D3DCOLOR_ARGB(0, R, G, B)
         else PDWORD (DWORD(d3dlr.pBits) + Y * dwDstPitch + X * 4)^ := D3DCOLOR_ARGB(255, R, G, B);
  end;

  hRet := FD3Texture.UnlockRect(0);
  if FAILED(hRet) then begin
     if Failed (hRet) then frmD3D.ErrorOut ('InitTexture', hRet);
     Exit;
  end;

  Bmp.Free;
  wrkBmp.Free;
end;

destructor TFish.Destroy;
begin
  if Assigned (FD3Texture) then begin
     FD3Texture._Release;
     FD3Texture := nil;
  end;
  inherited;
end;

function TFish.Draw : HRESULT;
var
  Vertices : ^TCustomVertex;
  hRet : HRESULT;
begin
  hRet := frmD3D.FD3DVB.Lock(0, 4 * SizeOf(TCustomVertex), PByte(Vertices), 0);
  if Failed(hRet) then begin
     Result := hRet;
     Exit;
  end;

  Vertices.X := -0.5 + PosX;
  Vertices.Y := -0.5 + PosY;
  Vertices.Z := 0;
  Vertices.U := 0;
  Vertices.V := 0;
  Inc(Vertices);

  Vertices.X := -0.5 + PosX;
  Vertices.Y := -0.5 + Scale + PosY;
  Vertices.Z := 0;
  Vertices.U := 0;
  Vertices.V := 1;
  Inc(Vertices);

  Vertices.X := -0.5 + Scale + PosX;
  Vertices.Y := -0.5 + PosY;
  Vertices.Z := 0;
  Vertices.U := 1;
  Vertices.V := 0;
  Inc(Vertices);

  Vertices.X := -0.5 + Scale + PosX;
  Vertices.Y := -0.5 + Scale + PosY;
  Vertices.Z := 0;
  Vertices.U := 1;
  Vertices.V := 1;

  frmD3D.FD3DVB.Unlock;

  with frmD3D.FD3DDevice do begin
       SetTexture(0, FD3Texture);
       SetTextureStageState(0, D3DTSS_COLOROP, D3DTA_TEXTURE);
       SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTA_TEXTURE);
       SetRenderState(D3DRS_ALPHABLENDENABLE, DWORD (True));
       SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
       SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
       DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
       SetRenderState(D3DRS_ALPHABLENDENABLE, DWORD (False));
  end;
  Result := frmD3D.FD3DDevice.SetTexture(0, nil);
end;

function TFish.RotateTexture : HRESULT;
var
  d3dlr : TD3DLOCKED_RECT;
  dwDstPitch : DWORD;
  pDst, pDst1 : PDWORD;
  X, Y : DWORD;
  wrkDW : DWORD;
begin
  FD3Texture.LockRect(0, d3dlr, nil, 0);

  dwDstPitch := d3dlr.Pitch;
  for Y := 0 to 127 do
        for X := 0 to 63 do begin
         pDst := PDWORD (DWORD(d3dlr.pBits) + Y * dwDstPitch + X * 4);
         pDst1 := PDWORD (DWORD(d3dlr.pBits) + Y * dwDstPitch + (127 - X) * 4);
         wrkDW := pDst^;
         pDst^ := pDst1^;
         pDst1^ := wrkDW;
  end;

  Result := FD3Texture.UnlockRect(0);
end;

procedure TFish.Move;
begin
  PosX := PosX + StepX;
  if (PosX < -1.5) or (PosX > 1.5) then begin
      RotateTexture;
      StepX := -StepX;
  end;
end;

end.
